
! (c) sparslab module name=bif_nu
!
! purpose:
!  a sample driver for computation of the nonsymmetric
!  incomplete decomposition BIF.
!  this driver does the following:
!    1) takes a simple matrix stored in csr scheme
!    2) computes nonsymmetric bif decomposition
!       in particular, on output we get factors L and U of the
!       LU decomposition
!    3) applies m to a given vector x of all ones
!
! history:
!   original version for sparslab - tu - 15/10/2008.
!
      use ainv90
!
      implicit none
      integer n,i
      double precision zero
      parameter(zero=0.0d0)
      integer size_u,size_v,size_l
      integer size_ut,size_vt,size_uu
      integer fillut,fillvt
      integer, allocatable :: il(:),jl(:)
      double precision, allocatable :: al(:)
      integer, allocatable :: iw(:),jw(:)
      double precision, allocatable :: aw(:)
      integer fillu,fillv,filll,filluu
      integer imodif,lsize,nx,imodift
      integer outlvl,ounit,info,ind
      integer fill,nja,ierr,mformat
      double precision diagtol,dshift,drflu,drflv,drfll,anorm
      double precision dvmaxval
      integer, allocatable, dimension(:) :: ia,ja
      double precision, allocatable, dimension(:) :: aa
      integer, allocatable, dimension(:) :: ib,jb,im,jm,dm
      double precision, allocatable, dimension(:) :: ab,am
      integer, allocatable, dimension(:) :: wn01,wn02
      double precision, allocatable, dimension(:) :: x,y,wr01
!
! start of bif_nu
!
!
!  -- generate input test matrix in csr format
!
      nx=37
      n=nx*nx
      nja=(4*2+(nx-2)*(nx-2)*4+4*(nx-2)*3)+n
      allocate(ia(n+1),stat=ierr)
      allocate(ja(nja),aa(nja),stat=ierr)
      call laplace(nx,ia,ja,aa)
!
!  -- unsymmetrize the matrix
!
      aa(2)=-2
      aa(3)=-2
!
!  -- get sample right-hand side
!
      allocate(x(n),y(n))
      do i=1,n
        x(i)=dble(i)
      end do
      call mvucsr1(n,ia,ja,aa,x,y)
!
!     %%%%%%%%%%%%%%%%%
!  -- to be set by user
!     %%%%%%%%%%%%%%%%%
!
!  -- set the drop tolerances and parameter lsize which determines
!  -- space for row pointers
!  -- here we set drop tolerances for zero to check the solution
!  -- but they can be set to, e.g., 0.1 or 0.01 etc.
!
      drflu=0.0d0
      drflv=0.0d0
      drfll=0.0d0
!
      lsize=min(100,n)
!
!  -- this is the parameter s from the SISC description
!  -- we will not modify it before calling nbif
!
      dshift=1.0d0
!
!  -- this means using preconditioner M_I from SISC
!  -- description
!
!     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!  -- end of the part to be set by user
!     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!
!
!     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!  -- start of defaults -- need not to be changed
!     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!
!  -- sizes: an initial first guess
!  -- will be automatically adapted inside the code
!
      size_u=1000000
      size_v=1000000
      size_ut=1000000
      size_vt=1000000
!
!  -- other parameters without the need to be changed
!  -- note that anorm: norm of A should be computed in advance
!  -- currently, it is not used inside the subroutine
!
      diagtol=1.1d-16
      anorm=10.
!
!     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!  -- end of defaults -- need not to be changed
!     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!
!
!  -- standard level of output
!
      outlvl=2
!
!  -- open unit for output
!
      ounit=2
      open(unit=ounit,file='output.out')
!
!  -- get the infinity matrix norm
!  -- get the shift from shift factor
!
      allocate(wr01(n),stat=ierr)
      if(ierr.ne.0) stop ': memory error'
      nja=ia(n+1)-1
      call csrnorms(n,n,ia(1),ja(1),aa(1),1,wr01)
      anorm=dvmaxval(n,wr01)
      deallocate(wr01)
!
!  -- compute A^T
!
      allocate(wn01(n+1),wn02(n+1),wr01(n),stat=ierr)
      if(ierr.ne.0) stop ': memory error'
      nja=ia(n+1)-1
      allocate(ib(n+1),jb(nja),ab(nja),stat=ierr)
      if(ierr.ne.0) stop ': memory error'
      call csr2csc1(n,ia,ja,aa,ib,jb,ab,wn01,wn02,0)
      deallocate(wr01,wn01,wn02)
!
!  -- left-looking nbifu
!  -- interleaved
!  -- version experiments_2008
!
      call nbif_interleave_experiments_2008
     *  (outlvl,ounit,n,ia,ja,aa,ib,jb,ab,
     *  il,jl,al,iw,jw,aw,size_u,size_v,
     *  size_ut,size_vt,
     *  size_l,size_uu,dshift,anorm,
     *  drflu,drflv,lsize,imodif,imodift,fillu,fillv,
     *  fillut,fillvt,filll,filluu,
     *  drfll,drfll,info)
!
      deallocate(ia,ja,aa,ib,jb,ab)
!
!  -- put the factors together (just for our convenience since
!  -- we have a corresponding solve routine)
!
      ind=il(n+1)-1+iw(n+1)-1-n
      fill=filll+filluu-n
      allocate(im(n+1),jm(ind),dm(n),am(ind),stat=ierr)
      if(ierr.ne.0) stop ': memory error'
      call cslu1(n,il(1),jl(1),al(1),
     *  iw(1),jw(1),aw(1),im(1),jm(1),dm(1),am(1))
!
!  -- output preconditioner info
!
      write(2,*)
      write(2,*)' preconditioner successfully computed'
      write(2,*)
      write(2,*)' %%%%%%%%%%%%%%%%%%%%%%%%%%%%'
      write(2,*)' the preconditioner'
      write(2,*)' %%%%%%%%%%%%%%%%%%%%%%%%%%%%'
      write(2,*)' drop tolerance for L',drfll
      write(2,*)' size of nbif factor L',filll
      write(2,*)' pointers for L'
      write(2,*) il(1:n+1)
      write(2,*)' column indices for L'
      write(2,*) jl(1:il(n+1)-1)
      write(2,*)' nonzero entries for L'
      write(2,*) al(1:il(n+1)-1)
      write(2,*)' drop tolerance for U',drfll
      write(2,*)' size of nbif factor U',filluu
      write(2,*)' pointers for U'
      write(2,*) iw(1:n+1)
      write(2,*)' column indices for U'
      write(2,*) jw(1:iw(n+1)-1)
      write(2,*)' nonzero entries for U'
      write(2,*) aw(1:iw(n+1)-1)
      deallocate(il,jl,al)
      deallocate(iw,jw,aw)
!
!  -- apply inverse of m (computed aism preconditioner) to
!  -- a vector y computed from a given vector x of all ones
!
!
      call bklu1(n,im,jm,am,dm,y,x,0)
!
      deallocate(im,jm,am)
!
!  -- print solution x
!
      write(2,*)' solution x'
      write(2,*) x(1:n)
      deallocate(x,y)
!
      close(2)
      stop ' success'
 400  continue
      stop ' some error'
!
! end of bif_nu
!
      end

