
* (c) sparslab module name=iwset
*
* purpose:
*   set entries of an integer vector to ialpha.
*   unrolled code.
*
* history:
*   original version for sparslab - tu - 14/1/1997
*
* parameters:
*
      subroutine iwset(n,ialpha,ix,incx)
c
c globals
      integer n,ialpha,incx
      integer ix(*)
c
c locals
      integer i,ind,m,mp1
c
c functions
      intrinsic mod
c
c start of iwset
c
      if(n.le.0) return
      if(incx.ne.1) then
c
c incx.ne.1
        ind=1
        if(incx.lt.0) ind=(-n+1)*incx+1
        do i=1,n
          ix(ind)=ialpha
          ind=ind+incx
        end do
        return
      else
c
c incx.eq.1
        m=mod(n,7)
        if(m.ne.0) then
          do i=1,m
            ix(i)=ialpha
          end do
          if(n.lt.7) return
        end if
        mp1=m+1
        do i=mp1,n,7
          ix(i)=ialpha
          ix(i+1)=ialpha
          ix(i+2)=ialpha
          ix(i+3)=ialpha
          ix(i+4)=ialpha
          ix(i+5)=ialpha
          ix(i+6)=ialpha
        end do
        return
      end if
c
c end of iwset
c
      return
c
      end

* (c) sparslab module name=idamin
*
* purpose:
*   finds the index of entry with minimum absolute value.
*
* history:
*   original version for sparslab - tu - 20/10/2005.
*   blas-like subroutine: rewritten from idamax.
*
* parameters:
*
      integer function idamin(n,dx,incx)
c
c parameters
c
      integer n,incx
      double precision dx(*)
c
c internals
c
      double precision dmin
      integer i,ix
c
c start of idamin
c
      idamin = 0
      if( n.lt.1 .or. incx.le.0 ) return
      idamin = 1
      if(n.eq.1) return
      if(incx.ne.1) then
c
c  -- code for increment not equal to 1
c
      ix = 1
      dmin = dabs(dx(1))
      ix = ix + incx
      do i = 2,n
        if(dabs(dx(ix)).lt.dmin) then
          idamin = i
          dmin = dabs(dx(ix))
        end if
        ix = ix + incx
      end do
c
c  -- return
c
      return
      end if
c
c  -- code for increment equal to 1
c
      dmin = dabs(dx(1))
      do i = 2,n
        if(dabs(dx(i)).ge.dmin) cycle
        idamin = i
        dmin = dabs(dx(i))
      end do
c
c  -- return
c
      return
c
c end of idamin
c
      end

* (c) sparslab module name=csrnorms
*
* purpose:
*  1,2,infinity norms of rows/columns of the matrix
*  or its transpose provided in csr format.
*
* history:
*   original version for sparslab - tu - 21/10/2005.
*
* parameters:
*      m by n matrix, ia(m+1),ja(nnz),aa(nnz),norm(m) or norm(n)
*  ii  typenorm  kind of demanded norm
*                =0:  infinity norm of the matrix
*                =1:  1-norm of the matrix
*                =2:  2-norm of rows of the matrix
*                =10: infinity norm of columns of the matrix
*                =11: 1-norm of columns of the matrix
*                =12: 2-norm of columns of the matrix
*                =15:
*
      subroutine csrnorms(m,n,ia,ja,aa,typenorm,norm)
c
c parameters
c
      implicit none
      integer m,n,typenorm
      integer ia(*),ja(*)
      double precision aa(*),norm(*)
c
c internals
c
      integer i,j,k,jstrt,jstop
      double precision zero,one,temp,temp2
      parameter(zero=0.0d0,one=1.0d0)
c
c start of csrnorms
c
      jstrt=ia(1)
      if(typenorm.eq.0) then
        do i=1,m
          jstop=ia(i+1)
          temp=zero
          do j=jstrt,jstop-1
            temp=temp+abs(aa(j))
          end do
          norm(i)=temp
          jstrt=jstop
        end do
      elseif(typenorm.eq.1) then
        do i=1,n
          norm(i)=zero
        end do
        do i=1,m
          jstop=ia(i+1)
          temp=zero
          do j=jstrt,jstop-1
            k=ja(j)
            norm(k)=norm(k)+abs(aa(j))
          end do
          jstrt=jstop
        end do
      elseif(typenorm.eq.2) then
        do i=1,m
          jstop=ia(i+1)
          temp2=zero
          do j=jstrt,jstop-1
            temp2=max(temp2,abs(aa(j)))
          end do
          if(temp2.le.zero) then
            norm(i)=temp2
          else
            do j=jstrt,jstop-1
              temp=temp+(aa(j)/temp2)**2
            end do
            temp=temp2*sqrt(temp)
            norm(i)=temp
          end if
          jstrt=jstop
        end do
      elseif(typenorm.eq.10) then
        do i=1,n
          norm(i)=zero
        end do
        do i=1,m
          jstop=ia(i+1)
          do j=jstrt,jstop-1
            k=ja(j)
            norm(k)=max(norm(k),abs(aa(j)))
          end do
          jstrt=jstop
        end do
      elseif(typenorm.eq.11) then
        do i=1,n
          norm(i)=zero
        end do
        do i=1,m
          jstop=ia(i+1)
          do j=jstrt,jstop-1
            norm(k)=norm(k)+abs(aa(j))
          end do
          jstrt=jstop
        end do
      elseif(typenorm.eq.12) then
c
c    -- dangerous strategy (a la sparskit)
c    -- for a safe procedure we need an allocatable or
c    -- to pass a real vector of the size n
c
        do i=1,n
          norm(i)=zero
        end do
        do i=1,m
          jstop=ia(i+1)
          do j=jstrt,jstop-1
            norm(k)=norm(k)+aa(j)*aa(j)
          end do
          jstrt=jstop
        end do
        do i=1,n
          norm(i)=sqrt(norm(i))
        end do
      end if
c
c  -- return
c
      return
c
c end of csrnorms
c
      end

* (c) sparslab module name=dvmaxval
*
* purpose:
*   maximum value of a component of a vector.
*
* history:
*   original version for sparslab - tu - 21/10/2005.
*
* parameters:
*  ii  n  vector dimension.
*  ri  x(n)  input vector.
*
      double precision function dvmaxval(n,x)
c
c parameters
c
      implicit none
      integer n
      double precision x(n)
c
c internals
c
      integer i
      double precision temp,xi
      double precision zero
      parameter (zero=0.0d0)
c
c start of dvmaxval
c
      temp=zero
      do i=1,n
        xi=x(i)
        temp=max(temp,xi)
      end do
      dvmaxval=temp
c
c  -- return
c
      return
c
c end of dvmaxval
c
      end

* (c) sparslab module name=csr2csc1
*
* purpose:
*  matrix format transformation routine: css,csr --> csc
*                                    or: css,csc --> csr
*
* history:
*   original version for sparslab - tu - 11/02/96
*
* parameters:
*
      subroutine csr2csc1(n,ia,ja,a,im,jm,am,wn01,wn02,mtyp)
c
c parameters
c
      integer n,mtyp
      integer ia(*),ja(*),im(*),jm(*)
      integer wn01(*),wn02(*)
      double precision a(*),am(*)
c
c internals
c
      integer i,j,k,l,jstrt,jstop
      integer izero
      parameter(izero=0)
c
c start of csr2csc1
c
c
c  -- initialization
c
      call iwset(n,izero,wn01,1)
c
c  -- find row counts
c
      do i=1,n
        jstrt=ia(i)
        jstop=ia(i+1)-1
        do j=jstrt,jstop
          k=ja(j)
          wn01(k)=wn01(k)+1
          if(mtyp.eq.1.and.i.ne.k) wn01(i)=wn01(i)+1
        end do
      end do
c
c  -- get partial sums and new pointers
c
      l=1
      wn02(1)=1
      im(1)=1
      do i=2,n+1
        l=l+wn01(i-1)
        wn02(i)=l
        im(i)=l
        wn01(i-1)=0
      end do
c
c  -- rewrite the structure
c
      do i=1,n
        jstrt=ia(i)
        jstop=ia(i+1)-1
        do j=jstrt,jstop
          k=ja(j)
          l=wn02(k)
          jm(l)=i
          am(l)=a(j)
          wn02(k)=wn02(k)+1
          if(mtyp.eq.1.and.i.ne.k) then
            l=wn02(i)
            jm(l)=k
            am(l)=a(j)
            wn02(i)=wn02(i)+1
          end if
        end do
      end do
c
c  -- return
c
      return
c
c end of csr2csc1
c
      end

* (c) sparslab module name=cslu1
*
* purpose:
*   Add L^T and U into one data structure for L/U.
*   Uses diagonal from L.
*   Matrices on input in CSS formats.
*
* history:
*   original version for sparslab - tu - 29/10/2007.
*
* parameters:
*
      subroutine cslu1(n,il,jl,al,iu,ju,au,ilu,jlu,dlu,alu)
c
c parameters
c
      implicit none
      integer n,il(*),jl(*),iu(*),ju(*)
      integer ilu(*),jlu(*),dlu(*)
      double precision al(*),au(*),alu(*)
c
c internals
c
      integer i,j,ind,indold,ierr
      integer, allocatable,dimension(:) :: list,start
      integer lstrt,lstop,ustrt,ustop,jndiag
      integer colnext,colcurr,nextrow
c
c start of cslu1
c
c  -- allocation
c
      allocate(list(n+1),start(n+1),stat=ierr)
      if(ierr.ne.0) call serr(0,2,'sparslab','cslu1',
     *  ' allocation error',8,2)
c
c  -- initialize working vectors
c
      call iwset(n,0,list(1),1)
c
c  -- initialize data structure for ILU
c
      ind=1
      indold=ind
      ilu(1)=ind
c
c  -- loop for rows
c
      do j=1,n
c
c    -- next column index of L
c
        colnext=list(j)
 200
     *  continue
        colcurr=colnext
        if(colcurr.ne.0) then
          colnext=list(colcurr)
          jndiag=start(colcurr)
          alu(ind)=al(jndiag)
          jlu(ind)=colcurr
          ind=ind+1
          lstrt=jndiag+1
          lstop=il(colcurr+1)-1
c
c      -- end of column of L
c
          if(lstrt.le.lstop) then
            start(colcurr)=lstrt
            nextrow=jl(lstrt)
            list(colcurr)=list(nextrow)
            list(nextrow)=colcurr
          end if
c
c      -- back to next column having a nonzero in a row i
c
          go to 200
c
        end if
c
c    -- add the row from U
c
        dlu(j)=ind
c
        ustrt=iu(j)
        ustop=iu(j+1)-1
        do i=ustrt,ustop
          alu(ind)=au(i)
          jlu(ind)=ju(i)
          ind=ind+1
        end do
c
c    -- set the end pointer
c
        ilu(j+1)=ind
c
c    -- sort the j-th row
c
        call uxvsr4(ind-indold,jlu(indold),alu(indold))
c
        indold=ind
c
c    -- define start,list
c
        if(il(j+1).ne.il(j)+1) then
          start(j)=il(j)+1
          colnext=jl(il(j)+1)
          list(j)=list(colnext)
          list(colnext)=j
        end if
c
      end do
c
c  -- deallocation
c
      deallocate(start,list)
c
c  -- return
c
      return
c
c end of cslu1
c
      end



! (c) sparslab module name=bklu1
!
! purpose:
!  matrix-vector multiplication for
!  factorized matrix format ilu1 (format 211).
!  used as nonsymmetric substitution for ilu(0) et al.
!
! parameters :
!  ii  job  option. if job=0 then y:=(a+e)**(-1)*x.
!                   if job=1 then y:=l**(-1)*x.
!                   if job=-1 then y:=u**(-1)*x
!                   if job=2 then y:=d**(-1)*x
! history:
!   original version for sparslab - tu - 13/09/1996
!   based on uxsgbb.f in naig.f et al.
!
! parameters:
!
      subroutine bklu1(n,ia,ja,a,id,x,y,job)
!
! parameters
!
      integer ia(*),ja(*),id(*),job
      double precision a(*),x(*),y(*)
      integer n
!
! internals
!
      integer i,j,k
      double precision temp,zero
      parameter(zero=0.0d0)
!
! start of bklu1
!
!  -- first phase: y=l^{-1}*x
!
      if(job.eq.0.or.job.eq.1) then
        do i=1,n
          temp=x(i)
          do j=ia(i),id(i)-1
            k=ja(j)
            temp=temp-y(k)*a(j)
          end do
          y(i)=temp
        end do
      end if
!
!  -- second phase: y=d^{-1}*y
!
      if(job.eq.0) then
        do i=1,n
          y(i)=y(i)*a(id(i))
        end do
      elseif(job.eq.2) then
        do i=1,n
          y(i)=x(i)*a(id(i))
        end do
      end if
!
!  -- third phase: y=u^{-1}*x
!
      if(job.eq.0.or.job.eq.-1) then
        if(job.eq.-1)  call dcopy(n,x,1,y,1)
        do i=n,1,-1
          temp=zero
          do j=id(i)+1,ia(i+1)-1
            k=ja(j)
            temp=temp+a(j)*y(k)
          end do
          temp=temp*a(id(i))
          y(i)=y(i)-temp
        end do
      end if
!
!  -- return
!
      return
!
! end of bklu1
!
      end subroutine bklu1


! (c) sparslab module name=serr
!
! purpose:
!  error output routine.
!
! history:
!   original version for sparslab - tu - 05/10/1997.
!
! parameters :
!  ii  eunit  error unit number.
!  ii  lunit  log unit number.
!  ci  libnam  package name.
!  ci  filnam  routine name.
!  ci  string  output string.
!  ii  err_no  error identifier.
!  ii  err_lev  error level:
!        =-1,0  for warnings; =1 for recoverable errors; =2 for fatal
!        errors.
!
      subroutine serr(eunit,lunit,libnam,filnam,string,err_no,err_lev)
!
! parameters
!
      integer eunit,lunit,err_no,err_lev
      integer len_output_string_max
      parameter(len_output_string_max=100)
!      character*(*) libnam,filnam,string
!      character*(*) libnam,filnam,string
!      character*(len(libnam)) :: libnam
!      character*(len(filnam)) :: filnam
!      character*(len(string)) :: string
      character (len=*) libnam,filnam,string
!      character (len_output_string_max) libnam,filnam,string
!
! functions
!
      integer lenstr0
!
! start of serr
!

      if(eunit.eq.0) then
        write(*,1) '       error no. ',err_no
        write(*,3) ' -- ',string(:lenstr0(string))
        write(*,2) '         found in routine  ',
     *    filnam(:lenstr0(filnam))
        write(*,2) '         called within package  ',
     *    libnam(:lenstr0(libnam))
      else
        write(eunit,1) '       error no. ',err_no
        write(eunit,3) ' -- ',string(:lenstr0(string))
        write(eunit,2) '         found in routine  ',
     *    filnam(:lenstr0(filnam))
        write(eunit,2) '         called within package  ',
     *    libnam(:lenstr0(libnam))
      end if
      if(lunit.ne.eunit) then
        write(lunit,1) '       error no. ',err_no
        write(lunit,3) ' -- ',string(:lenstr0(string))
        write(lunit,2) '         found in routine  ',
     *    filnam(:lenstr0(filnam))
        write(lunit,2) '         called within package  ',
     *    libnam(:lenstr0(libnam))
      end if
      if(err_lev.ge.2) then
!
!    -- fatal error
!
        stop
      end if
!
!  -- formats
!
 1
     * format((a),i7)
 2
     * format((a),2x,(a))
 3
     * format(9x,(a),(a))
!
!  -- return
!
      return
!
! end of serr
!
      end

* (c) sparslab module name=iwset1
*
* purpose:
*   set entries of an integer vector to ialpha.
*   unrolled code.
*
* history:
*   original version for sparslab - tu - 14/1/1997
*
* parameters:
*
      subroutine iwset1(n,ialpha,ix,incx,from)
c
c globals
      integer n,ialpha,incx,from
      integer ix(*)
c
c locals
      integer i,ind,m,mp1
c
c functions
      intrinsic mod
c
c start of iwset1
c
      if(n.le.0) return
      if(incx.ne.1) then
c
c incx.ne.1
        ind=1
        if(incx.lt.0) ind=(-n+1)*incx+1
        do i=1,n
          ix(ind+from-1)=ialpha
          ind=ind+incx
        end do
        return
      else
c
c incx.eq.1
        m=mod(n,7)
        if(m.ne.0) then
          do i=1,m
            ix(i+from-1)=ialpha
          end do
          if(n.lt.7) return
        end if
        mp1=m+1
        do i=mp1,n,7
          ix(i+from-1)=ialpha
          ix(i+1+from-1)=ialpha
          ix(i+2+from-1)=ialpha
          ix(i+3+from-1)=ialpha
          ix(i+4+from-1)=ialpha
          ix(i+5+from-1)=ialpha
          ix(i+6+from-1)=ialpha
        end do
        return
      end if
c
c end of iwset1
c
      return
c
      end

* (c) sparslab module name=dvcopy_backward
*
* purpose:
*   copy an integer vector in backward order.
*
* history:
*   original version for sparslab - tu - 19/03/2012.
*
* parameters:
*
      subroutine dvcopy_backward(n,dx,dy,from,to)
c
c parameters
c
      integer n,from,to
      double precision dx(*),dy(*)
c
c internals
c
      integer i
c
c functions
c
      intrinsic mod
c
c start of dvcopy_backward
c
      if(n.le.0) return
      do i=n,1,-1
        dy(i+to-1)=dx(i+from-1)
      end do
c
c  -- return
c
      return
c
c end of dvcopy_backward
c
      end

* (c) sparslab module name=iwcopy
*
* purpose:
*   copy an integer vector.
*   unrolled code.
*
* history:
*   original version for sparslab - tu - 14/1/1998.
*   rewritten from diniti.
*
* parameters:
*
      subroutine iwcopy(n,ix,incx,iy,incy)
c
c parameters
c
      integer n,incx,incy,ix(*),iy(*)
c
c locals
      integer i,intx,inty,m,mp1
c
c functions
c
      intrinsic mod
c
c start of iwcopy
c
      if(n.le.0) return
      if(.not.(incx.eq.1.and.incy.eq.1)) then
c
c nonequal increments
        intx=1
        inty=1
        if(incx.lt.0) intx=(-n+1)*incx+1
        if(incy.lt.0) inty=(-n+1)*incy+1
        do i=1,n
          iy(inty)=ix(intx)
          intx=intx+incx
          inty=inty+incy
        end do
      else
c
c increments equal to 1
        m=mod(n,7)
        if(m.ne.0) then
          do i=1,m
            iy(i)=ix(i)
          end do
          if(n.lt.7) return
        end if
        mp1=m+1
        do i=mp1,n,7
          iy(i)=ix(i)
          iy(i+1)=ix(i+1)
          iy(i+2)=ix(i+2)
          iy(i+3)=ix(i+3)
          iy(i+4)=ix(i+4)
          iy(i+5)=ix(i+5)
          iy(i+6)=ix(i+6)
        end do
      end if
c
      return
c
c end of iwcopy
c
      end

* (c) sparslab module name=ivcopy_backward
*
* purpose:
*   copy an integer vector in backward order.
*
* history:
*   original version for sparslab - tu - 19/03/2012.
*
* parameters:
*
      subroutine ivcopy_backward(n,ix,iy,from,to)
c
c parameters
c
      integer n,from,to,ix(*),iy(*)
c
c internals
c
      integer i
c
c functions
c
      intrinsic mod
c
c start of ivcopy_backward
c
      if(n.le.0) return
      do i=n,1,-1
        iy(i+to-1)=ix(i+from-1)
      end do
c
c  -- return
c
      return
c
c end of ivcopy_backward
c
      end

* (c) sparslab module name=garclds
*
* purpose:
*   garbage collection of a matrix in ds* format.
*
* history:
*   original version for sparslab - tu - 13/4/1996.
*
* parameters:
*  ii  n  dimension of the matrix in ds* format.
*  mu  ia(n)/ja(ia(n)+ka(n)-1)/ka(n)/aa(ia(n)+ka(n)-1) updated
*        matrix in ds* format.
*  ii  max_a  size of the vectors ja/aa.
*  ii  ibeg  position of first entry in ia/ja/ka/aa.
*  io  iend  position of last entry in ia/ja/ka/aa.
*  ii  nit  step number (used for modifications of distance parameter).
*  ii  idist  type of the distribution pattern of the ds* format.
*
      subroutine garclds(n,ia,ja,ka,aa,max_a,
     *  ibeg,iend,nit,idist)
c
c parameters
c
      integer n,ibeg,iend,max_a,nit,idist
      integer ia(*),ka(*),ja(*)
      double precision aa(*)
c
c internals
c
      integer i,j,k,l,istrt,isize
      integer izero
      double precision temp
      parameter(izero=0)
c
c start of garclds
c
c  -- compute the minimum necessary size of the compressed output
c
      isize=0
      do i=1,n
        if(ka(i).ne.0) then
          k=ia(i)
          l=ja(k)
          ja(k)=-i
          ia(i)=l
          isize=isize+ka(i)
        else
          ia(i)=ibeg
        end if
      end do
c
c  -- compute distance parameter idist
c
      if(idist.eq.-1) then
c
c  -- -- original case ldist.eq.'a'
c
        idist=max(0,(max_a-1-isize)/(2*(n-nit+1)))
      elseif(idist.gt.0) then
c
c  -- -- original case ldist.eq.'p'
c
        idist=min(max(0,(max_a-1-isize)/(2*(n-nit+1))),idist)
      else
c
c  -- -- original case ldist.eq.'n'
c
        idist=0
      endif
c
c  -- the garbage collection
c
      istrt=ibeg
      i=ibeg
 300
     * continue
      k=ja(i)
c
c  -- -- find a start of some row/column
c
      if(k.lt.0) then
        temp=aa(i)
c
c  -- -- move the row/column
c
        do j=1,ka(-k)-1
          ja(istrt+j)=ja(i+j)
          aa(istrt+j)=aa(i+j)
        end do
c
c  -- -- move the head and redefine the starting pointer
c
        aa(istrt)=temp
        ja(istrt)=ia(-k)
        ia(-k)=istrt
c
c  -- -- redefine starting pointers of both the source and target
c
        istrt=istrt+ka(-k)
        i=i+ka(-k)
c
c  -- -- clear some space (typically not large)
c
        do j=1,min(i-istrt,idist)
          ja(istrt)=0
          istrt=istrt+1
        end do
      else
c
c  -- -- move further
c
        i=i+1
      end if
c
c  -- -- end of coded loop
c
      if(i.lt.iend) go to 300
c
c  -- clear up to the end
c
      call iwset(iend-istrt+1,izero,ja(istrt),1)
c
c  -- redefine end
c
      iend=istrt-1
c
c  -- return
c
      return
c
c end of garclds
c
      end

* (c) sparslab module name=garclds2
*
* purpose:
*   garbage collection of a graph in ds* format.
*
* history:
*   original version for sparslab - tu - 13/4/1996.
*
* parameters:
*  ii  n  dimension of the graph in ds* format.
*  mu  ia(n)/ja(ia(n)+ka(n)-1)/ka(n) updated graph in ds* format.
*  ii  max_a  size of the vectors ja/aa.
*  ii  ibeg  position of first entry in ia/ja/ka/aa.
*  io  iend  position of last entry in ia/ja/ka/aa.
*  ii  idist  type of the distribution pattern of the ds* format.
*
      subroutine garclds2(n,ia,ja,ka,max_a,ibeg,iend,idist)
c
c parameters
c
      integer n,ibeg,iend,max_a,idist
      integer ia(*),ka(*),ja(*)
c
c internals
c
      integer i,j,k,l,istrt,isize
      integer izero
      parameter(izero=0)
c
c start of garclds2
c
c  -- compute the minimum necessary size of the compressed output
c
      isize=0
      do i=1,n
        if(ka(i).ne.0) then
          k=ia(i)
          l=ja(k)
          ja(k)=-i
          ia(i)=l
          isize=isize+ka(i)
        else
          ia(i)=ibeg
        end if
      end do
c
c  -- compute distance parameter idist
c
      if(idist.eq.-1) then
c
c  -- -- original ldist.eq.'a'
c
        idist=max(0,(max_a-1-isize)/(2*(n+1)))
      elseif(idist.gt.0) then
c
c  -- -- original ldist.eq.'p'
c
        idist=min(max(0,(max_a-1-isize)/(2*(n+1))),idist)
      else
c
c  -- -- original case ldist.eq.'n'
c
        idist=0
      endif
c
c  -- the garbage collection
c
      istrt=ibeg
      i=ibeg
 300
     * continue
      k=ja(i)
c
c  -- -- find a start of some row/column
c
      if(k.lt.0) then
c
c  -- -- move the vertex
c
        do j=1,ka(-k)-1
          ja(istrt+j)=ja(i+j)
        end do
c
c  -- -- move the head and redefine the starting pointer
c
        ja(istrt)=ia(-k)
        ia(-k)=istrt
c
c  -- -- redefine starting pointers of both the source and target
c
        istrt=istrt+ka(-k)
        i=i+ka(-k)
c
c  -- -- clear some space (typically not large)
c
        do j=1,min(i-istrt,idist)
          ja(istrt)=0
          istrt=istrt+1
        end do
      else
c
c  -- -- move further
c
        i=i+1
      end if
c
c  -- -- end of coded loop
c
      if(i.lt.iend) go to 300
c
c  -- clear up to the end
c
      call iwset(iend-istrt+1,izero,ja(istrt),1)
c
c  -- redefine end
c
      iend=istrt-1
c
c  -- return
c
      return
c
c end of garclds2
c
      end

! (c) sparslab module name=uxvsr4
!
! purpose:
!   ascending shellsort of an integer vector.
!   additional real vector is permuted in the same way.
!
! history:
!   original version for sparslab - tu - 15/01/1998.
!
! parameters:
!  ii  n  input vector length.
!  iu  arrayi(n)  input integer vector to be sorted.
!  ru  arrayr(n)  input real vector to be permuted in the same way.
!
      subroutine uxvsr4(k,arrayi,arrayr)
!
! parameters
!
      implicit none
      integer k
      integer arrayi(*)
      double precision arrayr(*)
!
! internals
!
      integer is,la,lt,ls,lls,i,j,js,khalf,ii
      double precision lb
!
! functions
!
      integer mylog2
!
! start of uxvsr4
!
!  -- nothing to be sorted
!
      if(k.le.1) return
!
!  -- shellsort initialization
!
      khalf=k/2
      ii=mylog2(2*k)
      ls=2**ii-1
!
!  -- shellsort loop
!
      do lt=1,ii
        if(ls.gt.khalf) then
          ls=ls/2
        else
          lls=k-ls
          do 200 i=1,lls
            is=i+ls
            la=arrayi(is)
            lb=arrayr(is)
            j=i
            js=is
 100
     *      continue
            if(la.ge.arrayi(j)) then
              arrayi(js)=la
              arrayr(js)=lb
              go to 200
            else
              arrayi(js)=arrayi(j)
              arrayr(js)=arrayr(j)
              js=j
              j=j-ls
            end if
            if(j.ge.1) go to 100
            arrayi(js)=la
            arrayr(js)=lb
 200
     *  continue
        ls=ls/2
        end if
      end do
!
!  -- return
!
      return
!
! end of uxvsr4
!
      end subroutine uxvsr4



! (c) sparslab module name=err
!
! purpose:
!  sparslab error handler
!
! history:
!   original version for sparslab - tu - 31/10/2003.
!
! parameters :
!  ii  eunit  error unit number.
!  ci  name  routine name.
!  ci  string  output string.
!  ii  err_no  error identifier.
!  ii  err_lev  error level:
!        =-1,0  for warnings; =1 for recoverable errors; =2 for fatal
!        errors.
!
      subroutine err(eunit,errtyp,name,err_lev)
!
! parameters
!
      use iotypes
      implicit none
      integer eunit,errtyp,err_lev
!      integer len_output_string_max
      character (len=*) name
!
! internals
!
      integer i,ii
      integer iii
      character err_desc*100
!
! functions
!
      integer lenstr,lenstr2
!
! start of err
!
      if(eunit.ne.0) then
        write(eunit,*)
        call errdesc(errtyp,err_desc)
        i=lenstr2(err_desc)
        write(eunit,1) ' error no. ',errtyp
        write(eunit,*) '  -- ',err_desc(:i)
        ii=lenstr(name)
        write(eunit,*) '  found in routine  ',name(:ii)
        write(eunit,*) '  called within the package -- sparslab --'
      end if
      write(*,*)
      call errdesc(errtyp,err_desc)
      i=lenstr2(err_desc)
      write(*,1) ' error no. ',errtyp
      write(*,*) '  -- ',err_desc(:i)
      ii=lenstr(name)
      write(*,*) '  found in routine  ',name(:ii)
      write(*,*) '  called within the package -- sparslab --'
!
      if(err_lev.eq.20) then ! not used tracing at this moment
 100
     *  continue
        iii=lenstr(trace%name)
        write(*,*) '-- called from  ',trace%name(:iii)
        if(trace%level.gt.1) then
          trace=>trace%prev
          go to 100
        end if
      end if
!
      if(err_lev.ge.2) then
!
!    -- fatal error
!
        stop
!
      end if
!
!  -- formats
!
 1
     * format((a),i7)
!
!  -- return
!
      return
!
! end of err
!
      end

* (c) sparslab module name=lenstr0
*
* purpose:
*  get length of string until character \0
*
* parameters :
*  ci  string  input string.
*  io  lenstr0  length of the string.
*
* history:
*   original version for sparslab - tu - 11/11/1997
*
      integer function lenstr0(string)
c
c parameters
c
      integer len_output_string_max
      parameter(len_output_string_max=100)
      character*(len_output_string_max) string
c
c internals
c
      integer i
c
c start of lenstr0
c
      i=index(string,'\0') - 1
      if(i.eq.-1) then
        i=len(string)
        i=1
      end if
c
      lenstr0=i
c
c  -- return
c
      return
c
c end of lenstr0
c
      end


* (c) sparslab module name=lenstr
*
* purpose:
*  get length of string.
*
* parameters :
*  ci  string  input string.
*  io  lenstr  length of the string.
*
* history:
*   original version for sparslab - tu - 11/11/1997
*
      integer function lenstr(string)
c
c parameters
c
      integer len_output_string_max
      parameter(len_output_string_max=100)
      character*(len_output_string_max) string
c
c internals
c
      integer i
c
c start of lenstr
c
      i=index(string,' ') - 1
      if(i.eq.-1) then
        i=len(string)
        i=1
      end if
c
      lenstr=i
c
c  -- return
c
      return
c
c end of lenstr
c
      end


* (c) sparslab module name=lenstr2
*
* purpose:
*  get length of string.
*  function trying to find the string lengths from back
*  assuming maximal length equal to len_output_string_max.
*  used for string outputs.
*
* parameters :
*  ci  string  input string.
*  io  lenstr2  length of the string.
*
* history:
*   original version for sparslab - tu - 14/9/1998
*
      integer function lenstr2(string)
c
c parameters
c
      integer len_output_string_max
      parameter(len_output_string_max=100)
      character*(len_output_string_max) string
c
c locals
      integer i
c
c start of lenstr2
c
      do i=len_output_string_max,1,-1
        if(string(i:i).ne.' ') then
          lenstr2=i
          return
        end if
      end do
      lenstr2=0
c
      return
c
c end of lenstr2
c
      end


! (c) sparslab module name=mylog2
!
! purpose:
!   upper whole part of logaritm with base two.
!   the name is chosen in order to avoid conflicts
!   with some compilers and their intrinsic funcions.
!
! history:
!   original version for sparslab - tu - 14/02/2001.
!
! parameters:
!
      integer function mylog2(n)
!
! parameters
!
      integer n
!
! internals
!
      integer j,k
!
! start of mylog2
!
      k=0
      j=n
      if(n.gt.1) then
 100
     *  continue
        if(j.eq.1) go to 200
        k=k+1
        j=j/2+mod(j,2)
        go to 100
      end if
 200
     * continue
      mylog2=k
!
!  -- return
!
      return
!
! end of mylog2
!
      end


* (c) sparslab module name=dinit
*
* purpose:
*   initialization of a dense double precision vector by alpha
*
* history:
*   original version for sparslab - tu - 18/9/1996
*
      subroutine dinit(n,alpha,x,incx)
c
c parameters
c
      integer n,incx
      double precision alpha,x(*)
c
c internals
c
      integer i,ix,m,mp1
c
c functions
c
      intrinsic mod
c
c start of dinit
c
      if (n.le.0) return
      if(incx.ne.1) then
c
c    -- incx.ne.1
c
        ix=1
        if(incx.lt.0) ix=(-n+1)*incx+1
        do i=1,n
          x(ix)=alpha
          ix=ix+incx
        end do
        return
      else
c
c    -- incx.eq.1
c
        m=mod(n,7)
        if(m.ne.0) then
          do i=1,m
            x(i)=alpha
          end do
          if(n.lt.7) return
        end if
        mp1=m+1
        do i=mp1,n,7
          x(i)=alpha
          x(i+1)=alpha
          x(i+2)=alpha
          x(i+3)=alpha
          x(i+4)=alpha
          x(i+5)=alpha
          x(i+6)=alpha
        end do
        return
      end if
c
c end of dinit
c
      end

! (c) sparslab module name=srtcs1
!
! purpose:
!   sort a sparse matrix stored in a cs storage format.
!
! history:
!   original version for sparslab - tu - 01/06/1996.
!
! parameters:
!   ii  n  matrix dimension.
!   ou  ia(n+1)/ja(ia(n+1)-1/a(ia(n+1)-1  updated matrix cs object.
!
      subroutine srtcs1(n,ia,ja,aa)
!
! parameters
!
      implicit none
      integer n,ia(*),ja(*)
      double precision aa(*)
!
! internals
!
      integer i,jlen,jstrt,jstop
!
! start of srtcs1
!
      jstop=ia(1)
      do i=1,n
        jstrt=jstop
        jstop=ia(i+1)
        jlen=jstop-jstrt
        call uxvsr4(jlen,ja(jstrt),aa(jstrt))
      end do
!
!  -- return
!
      return
!
! end of srtcs1
!
      end subroutine srtcs1


! (c) sparslab module name=laplace
!
! purpose:
!  generate a real square matrix.
!  simple 5pt discretization of 2D laplace equation.
!
!  -\nabla(u)
!
! Its entries are:
!
!   4    -1  ..... -1 ....
!  -1   4  ..................
!    .......................................
!  -1    ............................
!
!
! history:
!   original version for sparslab - tu - 07/05/2008.
!
      subroutine laplace(nx,ia,ja,aa)
!
! parameters
!
      implicit none
      integer nx,n
      integer :: ia(*),ja(*)
      double precision :: aa(*)
!
!
! start of laplace
!
!  -- Laplace on 2d square grid, dirichlet bc
!
      n=nx*nx
      call grid5pt(nx,n,ia,ja)
      call val5ptb(n,ia,ja,aa,nx)
!
!  -- return
!
      return
!
! end of laplace
!
      end subroutine laplace



! (c) sparslab module name=grid5pt
!
! purpose:
!   get matrix from the 5pt stencil.
!
! history:
!   original version for sparslab - tu - 18/7/1998.
!
! parameters:
!
      subroutine grid5pt(size,n,ia,ja)
!
      integer size,n
      integer ia(*),ja(*)
!
! internals
!
      integer i,j,k,next
!
! start of grid5pt
!
      next=1
      k=1
      do i=1,size
        do j=1,size
          ia(k)=next
          if(i.ge.2) then
            ja(next)=k-size
            next=next+1
          end if
          if(j.ge.2) then
            ja(next)=k-1
            next=next+1
          end if
          if(j.lt.size) then
            ja(next)=k+1
            next=next+1
          end if
          if(i.lt.size) then
            ja(next)=k+size
            next=next+1
          end if
          k=k+1
        end do
      end do
      ia(k)=next
      n=size*size
!
!  -- return
!
      return
!
! end of grid5pt
!
      end subroutine grid5pt


* (c) sparslab module name=val5ptb
*
* purpose:
*   put numerical values into the matrix with 5pt stencil.
*   uses the model:
*           [-u_xx-u_yy] = ...
*
* history:
*   original version for sparslab - tu - 18/7/1998.
*
* parameters:
*
      subroutine val5ptb(n,ia,ja,aa,nx)
c
c parameters
c
      implicit none
      integer n
      integer nx
      integer ia(*),ja(*)
      double precision aa(*)
c
c internals
c
      integer i,j,k,ind,jstrt,jstop,noffdiag,ideg
      double precision hstep
      double precision one,four
      parameter(one=1.0d0,four=4.0d0)
c
c start of val5ptb
c
      noffdiag=ia(n+1)-1
      ind=noffdiag+n
      do i=noffdiag,1,-1
        ja(ind)=ja(i)
        ind=ind-1
      end do
c
c  -- values
c
      hstep=one/(dble(nx)+one)
c
      ind=1
      jstop=n
      do i=1,n
        ia(i)=ind
        jstrt=jstop+1
        jstop=ia(i+1)-1+n
        ideg=jstop-jstrt+1
        do j=jstrt,jstop
          k=ja(j)
          if(k.lt.i) then
            ja(ind)=ja(j)
              if(k.eq.(i-1)) then
                aa(ind)=-one
              else
                aa(ind)=-one
              end if
            ind=ind+1
          else
            jstrt=j
            exit
          end if
        end do
        ja(ind)=i
        aa(ind)=four
        if(i.eq.n) jstrt=jstop+1
        ind=ind+1
        do j=jstrt,jstop
          k=ja(j)
          ja(ind)=ja(j)
            if(k.eq.(i+1)) then
              aa(ind)=-one
            else
              aa(ind)=-one
            end if
          ind=ind+1
        end do
      end do
      ia(n+1)=noffdiag+n+1
c
c  -- return
c
      return
c
c end of val5ptb
c
      end subroutine val5ptb






! (c) sparslab module name=errdesc
!
! purpose:
!   error / warning  descriptors.
!
! history:
!   original version for sparslab - tu - 31/10/2003.
!
! parameters:
!
      subroutine errdesc(errtyp,err_desc)
!
! parameters
!
      integer errtyp
      integer len_output_string_max
      parameter(len_output_string_max=100)
      character*(len_output_string_max) err_desc
!
! start of errdesc
!
      call stinit(err_desc)
      if(errtyp.eq.0) then
        err_desc=' sucess'
      elseif(errtyp.eq.1) then
        err_desc='dimension nmax too small'
      elseif(errtyp.eq.2) then
        err_desc='dimension mmax too small'
      elseif(errtyp.eq.3) then
        err_desc='matrix is not rectangular'
      elseif(errtyp.eq.4) then
        err_desc='error opening input file'
      elseif(errtyp.eq.5) then
        err_desc='error opening output file'
      elseif(errtyp.eq.6) then
        err_desc='error reading input file'
      elseif(errtyp.eq.7) then
        err_desc='error writing output file'
      elseif(errtyp.eq.8) then
        err_desc='allocation error'
      elseif(errtyp.eq.9) then
        err_desc='deallocation error'
      elseif(errtyp.eq.10) then
        err_desc='error closing file'
      elseif(errtyp.eq.11) then
        err_desc='error reading matrix input type'
      elseif(errtyp.eq.12) then
        err_desc='no space for the diagonal elements in the coox scheme'
      elseif(errtyp.eq.13) then
        err_desc='error reading preconditioning type'
      elseif(errtyp.eq.14) then
        err_desc='error opening matrix definition file slabmatr.df'
      elseif(errtyp.eq.15) then
        err_desc='error reading matrix definition file slabmatr.df'
      elseif(errtyp.eq.16) then
        err_desc='matrix has all elements equal to zero'
      elseif(errtyp.eq.17) then
        err_desc='matrix is not ordered'
      elseif(errtyp.eq.18) then
        err_desc='missing diagonal elements'
      elseif(errtyp.eq.20) then
        err_desc='error while reading data'
      elseif(errtyp.eq.21) then
        err_desc='real working space not sufficient'
      elseif(errtyp.eq.22) then
        err_desc='no space for dynamic data structures'
      elseif(errtyp.eq.23) then
        err_desc='matrix format not supported'
      elseif(errtyp.eq.24) then
        err_desc='matrix is structurally singular'
      elseif(errtyp.eq.25) then
        err_desc='no space for basis columns'
      elseif(errtyp.eq.26) then
        err_desc='matrix used in the linpack is not positive definite'
      elseif(errtyp.eq.27) then
        err_desc='null matrix'
      elseif(errtyp.eq.28) then
        err_desc='unknown norm type'
      elseif(errtyp.eq.29) then
        err_desc='unknown preconditioner type'
      elseif(errtyp.eq.30) then
        err_desc='dimension submax too small'
      elseif(errtyp.eq.31) then
        err_desc='dimension lnzmax too small'
      elseif(errtyp.eq.32) then
        err_desc='not a matrix from mixed-hybrid fem problem'
      elseif(errtyp.eq.33) then
        err_desc='dimension npmax too small'
      elseif(errtyp.eq.34) then
        err_desc='unknown matrix format type'
      elseif(errtyp.eq.35) then
        err_desc='this diag_one is not supported'
      elseif(errtyp.eq.36) then
        err_desc='bad computer setting'
      elseif(errtyp.eq.37) then
        err_desc='bad expl_lev level of expliciteness'
      elseif(errtyp.eq.38) then
        err_desc='min(pmax,mmax) is too small'
      elseif(errtyp.eq.39) then
        err_desc='matrix a is not partially ordered'
      elseif(errtyp.eq.40) then
        err_desc='number of processors greater than anticipated'
      elseif(errtyp.eq.41) then
        err_desc='number of blocks/processes greater than the workspace'
      elseif(errtyp.eq.42) then
        err_desc='unknown element type'
      elseif(errtyp.eq.43) then
        err_desc='unknown boundary condition'
      elseif(errtyp.eq.44) then
        err_desc='corrupted input tg files'
      elseif(errtyp.eq.45) then
        err_desc='modification type not supported'
      elseif(errtyp.eq.46) then
        err_desc='incompatible input data'
      elseif(errtyp.eq.47) then
        err_desc='dimension pmax too small'
      elseif(errtyp.eq.48) then
        err_desc='strange input matrix'
      elseif(errtyp.eq.49) then
        err_desc='this idgtype is not supported'
      elseif(errtyp.eq.50) then
        err_desc='this preconditioner option is not supported'
      elseif(errtyp.eq.51) then
        err_desc='maximum length of an array of variable size exceeded'
      elseif(errtyp.eq.52) then
        err_desc='maximum length of an array of variable size exceeded'
      elseif(errtyp.eq.53) then
        err_desc='shr_prec not supported'
      elseif(errtyp.eq.54) then
        err_desc='nonzero ini_app is not supported in this case'
      elseif(errtyp.eq.55) then
        err_desc='zero row in a matrix'
      elseif(errtyp.eq.56) then
        err_desc='dimension mmax is too small'
      elseif(errtyp.eq.57) then
        err_desc='dimension nzc is too small'
      elseif(errtyp.eq.58) then
        err_desc='the symmetric matrix is not symmetrized'
      elseif(errtyp.eq.61) then
        err_desc='bad input matrix dimension n'
      elseif(errtyp.eq.62) then
        err_desc='bad input number of matrix nonzeros nja (mmax)'
      elseif(errtyp.eq.63) then
        err_desc='small dimension nmax1 for column indices and nonzeros'
      elseif(errtyp.eq.64) then
        err_desc='small dimension for row indices and nonzeros mmax2'
      elseif(errtyp.eq.65) then
        err_desc='small dimension for short vectors nmax1'
      elseif(errtyp.eq.66) then
        err_desc='bad pv_pass'
      elseif(errtyp.eq.67) then
        err_desc='empty column in the initial matrix'
      elseif(errtyp.eq.68) then
        err_desc='empty row in the initial matrix'
      elseif(errtyp.eq.69) then
        err_desc='repeated matrix indices'
      elseif(errtyp.eq.70) then
        err_desc='matrix incorrect'
      elseif(errtyp.eq.201) then
        err_desc='interrupted basis update'
      elseif(errtyp.eq.202) then
        err_desc='not enough space for the preconditioner'
      elseif(errtyp.eq.203) then
        err_desc='lu factorization error'
      elseif(errtyp.eq.601) then
        err_desc='factorial overflow'
      elseif(errtyp.eq.602) then
        err_desc='binomial coefficients overflow'
      elseif(errtyp.eq.801) then
        err_desc='lapack dpotrf error'
      elseif(errtyp.eq.802) then
        err_desc='lapack dpotrf error'
      elseif(errtyp.eq.803) then
        err_desc='lapack dormqr error no.'
      elseif(errtyp.eq.804) then
        err_desc='lapack dgeqrf error no.'
      elseif(errtyp.eq.805) then
        err_desc='lapack dtptrs error no.'
      elseif(errtyp.eq.806) then
        err_desc='lapack dgetrf error no.'
      elseif(errtyp.eq.807) then
        err_desc='lapack dgetrs error no.'
!
!*********************
!****************************************************
! new list of errors
!
!  -- memory faults
!
      elseif(errtyp.eq.1001) then
        err_desc='matrix incorrect -- large indices found'
      elseif(errtyp.eq.1002) then
        err_desc='not a matrix from mixed-hybrid fem problem'
      elseif(errtyp.eq.1010) then
        err_desc='needed integer*8 at least'
!
!  -- format faults
!
      elseif(errtyp.eq.1031) then
        err_desc='matrix format not supported'
      elseif(errtyp.eq.1032) then
        err_desc='diagonal elements assumed'
      elseif(errtyp.eq.1033) then
        err_desc='matrix is not ordered'
      elseif(errtyp.eq.1034) then
        err_desc='matrix file does not contain what expected'
      elseif(errtyp.eq.1041) then
        err_desc='matrix is structurally singular'
      elseif(errtyp.eq.1045) then
        err_desc='matrix is singular and it was modified'
      elseif(errtyp.eq.1046) then
        err_desc='problems with positive definiteness'
      elseif(errtyp.eq.1051) then
        err_desc='matrix format error'
      elseif(errtyp.eq.1052) then
        err_desc='invalid matrix file format'
      elseif(errtyp.eq.1061) then
        err_desc='internal error of dynamic data structures'
      elseif(errtyp.eq.1101) then
        err_desc='allocated space is not sufficient'
      elseif(errtyp.eq.1102) then
        err_desc='space for dynamic data structures is not sufficient'
!
! graph partitioning
!
      elseif(errtyp.eq.1201) then
        err_desc='graph structure is not connected'
      elseif(errtyp.eq.1202) then
        err_desc='number of partitions is not a power of 2'
      elseif(errtyp.eq.1203) then
        err_desc='vertex partitioner is incorrect'
!
!  -- vector errors
!
      elseif(errtyp.eq.1301) then
        err_desc='invalid vector norm type'
!
!  -- direct method errors
!
      elseif(errtyp.eq.1401) then
        err_desc='symbolic factorization failed'
!
!  -- preconditioner computation errors
!
      elseif(errtyp.eq.1501) then
        err_desc='this preconditioning option is not supported'
      elseif(errtyp.eq.1502) then
        err_desc='error reading preconditioner type'
      elseif(errtyp.eq.1511) then
        err_desc='incorrect preconditioner pattern specification'
!
!  -- missing space for hardly predictable space demands
!
      elseif(errtyp.eq.1601) then
        err_desc='missing space to store givens rotations'
      elseif(errtyp.eq.1602) then
        err_desc='missing space to store null space basis'
      elseif(errtyp.eq.1603) then
        err_desc='missing space to store the preconditioner'
      elseif(errtyp.eq.1604) then
        err_desc='missing space to store elementary matrices'
      elseif(errtyp.eq.1605) then
        err_desc='allocation failure'
!
!  -- warnings
!
      elseif(errtyp.eq.1653) then
        err_desc='interrupted update of the preconditioner'
!
!  -- errors in incomplete decompositions
!
!
!  -- warnings
!
      elseif(errtyp.eq.1851) then
        err_desc='missing diagonal entries'
!
!  -- general utilities errors
!
      elseif(errtyp.eq.2001) then
        err_desc='arguments missing or invalid'
      elseif(errtyp.eq.2010) then
        err_desc='syntax error'
!
!  -- i/o errors
!
      elseif(errtyp.eq.2101) then
        err_desc='error opening input file'
      elseif(errtyp.eq.2102) then
        err_desc='error opening output file'
      elseif(errtyp.eq.2103) then
        err_desc='error opening file'
      elseif(errtyp.eq.2104) then
        err_desc='error reading matrix name'
      elseif(errtyp.eq.2105) then
        err_desc='error opening matrix file'
      elseif(errtyp.eq.2106) then
        err_desc='error reading a parameter'
      elseif(errtyp.eq.2107) then
        err_desc='error reading preconditioner type'
      elseif(errtyp.eq.2108) then
        err_desc='error reading right-hand side file'
      elseif(errtyp.eq.2109) then
        err_desc='error reading null-space basis type'
      elseif(errtyp.eq.2110) then
        err_desc='error reading smoother type'
      elseif(errtyp.eq.2111) then
        err_desc='error reading iterative method type'
!
      elseif(errtyp.eq.2201) then
        err_desc='unit does not correspond to an opened file'
!
!  -- sti preprocessor
!
      elseif(errtyp.eq.2501) then
        err_desc='sti syntax error'
!
!  -- text processing routines errors
!
      elseif(errtyp.eq.2651) then
        err_desc='symbol table full -- increase its dimension'
      elseif(errtyp.eq.2652) then
        err_desc='symbol not found in the table'
!
!  -- block methods
!
      elseif(errtyp.eq.3201) then
        err_desc='incorrect static block size'
!
!  -- linpack errors
!
      elseif(errtyp.eq.8001) then
        err_desc='linpack dpofa error'
      elseif(errtyp.eq.8002) then
        err_desc='linpack dpodi error'
!
!  -- lapack errors
!
      elseif(errtyp.eq.8101) then
        err_desc='lapack dpotrf error'
      elseif(errtyp.eq.8102) then
        err_desc='lapack dpotrs error'
!
!  -- sparslab team insufficiencies
!
      elseif(errtyp.eq.9900) then
        err_desc='sorry, this combination of inputs not yet implemented'
      end if
!
!  -- return
!
      return
!
! end of errdesc
!
      end


* (C) SPARSLAB MODULE NAME=STINIT
*
* PURPOSE:
*  INITIALIZE AN OUTPUT STRING OF A FIXED SIZE.
*
* HISTORY:
*   ORIGINAL VERSION FOR SPARSLAB - TU - 12/9/1998
*
* PARAMETERS :
*  CU  STRING  A STRING.
*
      SUBROUTINE STINIT(STRING)
C
C GLOBALS
      INTEGER LEN_OUTPUT_STRING_MAX
      PARAMETER(LEN_OUTPUT_STRING_MAX=100)
      CHARACTER*(LEN_OUTPUT_STRING_MAX) STRING
C
C LOCALS
      INTEGER I
C
C START OF STINIT
C
      DO I=1,LEN_OUTPUT_STRING_MAX
        STRING(I:I)=' '
      END DO
C
      RETURN
C
C END OF STINIT
C
      END


! (c) sparslab module name=mvucsr1
!
! purpose:
!   sparse matrix-vector multiplication y=a*x.
!   csr format.
!
! history:
!   original version for sparslab - tu - 01/05/1996.
!   modified - tu - 29/8/1997
!
! parameters:
!   ii  m  number of matrix rows.
!   mi  ia(m+1)/ja(ia(m+1)-1)/a(ia(m+1)-1)  input matrix csr object.
!   ri  x(n)  input vector.
!   ro  y(m)  output vector y=a*x.
!
      subroutine mvucsr1(m,ia,ja,aa,x,y)
!
! parameters
!
      integer m
      integer ia(*),ja(*)
      double precision aa(*),x(*),y(*)
!
! internals
!
      integer i,j,k,jstrt,jstop
      double precision temp,zero
      parameter(zero=0.0d0)
!
! start of mvucsr1
!
      jstrt=1
      do i=2,m+1
        jstop=ia(i)
        temp=zero
        do j=jstrt,jstop-1
          k=ja(j)
          temp=temp+aa(j)*x(k)
        end do
        y(i-1)=temp
        jstrt=jstop
      end do
!
!  -- return
!
      return
!
! end of mvucsr1
!
      end

